import {
  BASE_URL,
  YOUTUBE_CLIENT_ID,
  YOUTUBE_CLIENT_SECRET,
  YOUTUBE_PLAYLIST_ID
} from '$env/static/private';
import { log } from '$lib/log';
import type { OauthResponse } from '$lib/mastodon/response';
import type { SongInfo } from '$lib/odesliResponse';
import fs from 'fs/promises';

export class YoutubePlaylistAdder {
  private apiBase: string = 'https://www.googleapis.com/youtube/v3';
  private token_file_name: string = 'yt_auth_token';

  /// How many minutes before expiry the token will be refreshed
  private refresh_time: number = 15;

  public async authCodeExists(): Promise<boolean> {
    try {
      const fileHandle = await fs.open(this.token_file_name);
      await fileHandle.close();
      return true;
    } catch {
      log.info('No auth token yet, authorizing...');
      return false;
    }
  }

  public constructAuthUrl(redirectUri: URL): URL {
    const endpoint = 'https://accounts.google.com/o/oauth2/v2/auth';
    const authUrl = new URL(endpoint);
    authUrl.searchParams.append('client_id', YOUTUBE_CLIENT_ID);
    authUrl.searchParams.append('redirect_uri', redirectUri.toString());
    authUrl.searchParams.append('response_type', 'code');
    authUrl.searchParams.append('scope', 'https://www.googleapis.com/auth/youtube');
    authUrl.searchParams.append('access_type', 'offline');
    authUrl.searchParams.append('include_granted_scopes', 'false');
    return authUrl;
  }

  public async receivedAuthCode(code: string, url: URL) {
    log.debug('received code');
    const tokenUrl = new URL('https://oauth2.googleapis.com/token');
    const params = new URLSearchParams();
    params.append('client_id', YOUTUBE_CLIENT_ID);
    params.append('client_secret', YOUTUBE_CLIENT_SECRET);
    params.append('code', code);
    params.append('grant_type', 'authorization_code');
    params.append('redirect_uri', `${url.origin}${url.pathname}`);
    log.debug('sending token req', params);
    const resp: OauthResponse = await fetch(tokenUrl, {
      method: 'POST',
      body: params
    }).then((r) => r.json());
    log.debug('received access token', resp);
    let expiration = new Date();
    expiration.setTime(expiration.getTime() + resp.expires_in * 1000);
    expiration.setSeconds(expiration.getSeconds() + resp.expires_in);
    resp.expires = expiration;
    await fs.writeFile(this.token_file_name, JSON.stringify(resp));
  }

  private async auth(): Promise<OauthResponse | null> {
    try {
      const youtube_token_file = await fs.readFile(this.token_file_name, { encoding: 'utf8' });
      let token = JSON.parse(youtube_token_file);
      log.debug('read youtube access token', token);
      if (token.expires) {
        if (typeof token.expires === typeof '') {
          token.expires = new Date(token.expires);
        }
      }
      return token;
    } catch (e) {
      log.error('Could not read youtube access token', e);
      return null;
    }
  }

  private async refreshToken(): Promise<OauthResponse | null> {
    const token = await this.auth();
    if (token == null || !token?.expires) {
      return null;
    }
    let now = new Date();
    now.setTime(now.getTime() - this.refresh_time * 60 * 1000);
    log.info('token expiry', token.expires, 'vs refresh @', now);
    if (token.expires.getTime() > now.getTime()) {
      return token;
    }

    log.info(
      'YT token expires',
      token.expires,
      token.expires.getTime(),
      `which is less than ${this.refresh_time} minutes from now`,
      now,
      now.getTime()
    );

    const tokenUrl = new URL('https://oauth2.googleapis.com/token');
    const params = new URLSearchParams();
    params.append('client_id', YOUTUBE_CLIENT_ID);
    params.append('client_secret', YOUTUBE_CLIENT_SECRET);
    params.append('refresh_token', token.refresh_token || '');
    params.append('grant_type', 'refresh_token');
    params.append('redirect_uri', `${BASE_URL}/ytauth`);
    if (!token.refresh_token) {
      log.error('Need to refresh access token, but no refresh token provided');
      return null;
    }
    log.debug('sending token req', params);
    let resp: OauthResponse = await fetch(tokenUrl, {
      method: 'POST',
      body: params
    }).then((r) => r.json());
    if (resp.error) {
      log.error('token resp error', resp);
      return null;
    }
    if (!resp.refresh_token) {
      resp.refresh_token = token.refresh_token;
    }
    let expiration = new Date();
    expiration.setSeconds(expiration.getSeconds() + resp.expires_in);
    resp.expires = expiration;
    await fs.writeFile(this.token_file_name, JSON.stringify(resp));
    return resp;
  }

  public async addToPlaylist(song: SongInfo) {
    log.debug('addToYoutubePlaylist');
    const token = await this.refreshToken();
    if (token == null) {
      return;
    }

    if (!YOUTUBE_PLAYLIST_ID || YOUTUBE_PLAYLIST_ID === 'CHANGE_ME') {
      log.debug('no playlist ID configured');
      return;
    }
    if (!song.youtubeUrl) {
      log.debug('Skip adding song to YT playlist, no youtube Url', song);
      return;
    }

    const songUrl = new URL(song.youtubeUrl);
    const youtubeId = songUrl.searchParams.get('v');
    if (!youtubeId) {
      log.debug(
        'Skip adding song to YT playlist, could not extract YT id from URL',
        song.youtubeUrl
      );
      return;
    }
    log.debug('Found YT id from URL', song.youtubeUrl, youtubeId);

    const playlistItemsUrl = new URL(this.apiBase + '/playlistItems');
    playlistItemsUrl.searchParams.append('videoId', youtubeId);
    playlistItemsUrl.searchParams.append('playlistId', YOUTUBE_PLAYLIST_ID);
    playlistItemsUrl.searchParams.append('part', 'id');
    const existingPlaylistItem = await fetch(this.apiBase + '/playlistItems', {
      headers: { Authorization: `${token.token_type} ${token.access_token}` }
    }).then((r) => r.json());
    log.debug('existingPlaylistItem', existingPlaylistItem);
    if (existingPlaylistItem.pageInfo && existingPlaylistItem.pageInfo.totalResults > 0) {
      log.info('Item already in playlist');
      return;
    }

    const searchParams = new URLSearchParams([['part', 'snippet']]);
    const options: RequestInit = {
      method: 'POST',
      headers: { Authorization: `${token.token_type} ${token.access_token}` },
      body: JSON.stringify({
        snippet: {
          playlistId: YOUTUBE_PLAYLIST_ID,
          resourceId: {
            videoId: youtubeId,
            kind: 'youtube#video'
          }
        }
      })
    };
    const youtubeApiUrl = new URL(`${this.apiBase}/playlistItems?${searchParams}`);
    const resp = await fetch(youtubeApiUrl, options);
    const respObj = await resp.json();
    log.debug('Added to playlist', options, respObj);
    if (respObj.error) {
      log.debug('Add to playlist failed', respObj.error.errors);
    }
  }
}
